# The process_data.r script by Sharon Goldwater

# Used only for convenience while looking at data:
# index, ref_index, ID, hyp
# 
# Result variables:
# wi, wd, ws, eval, wd_or_ws, rer, is_err
# 
# Unused (obsolete):
# wlen_c, ref_length, ref_pos_f, ref_pos_b, is_first, is_last, is_only, pos, start_time, end_time, oqmin, oqmax, oqmean, noise, pause_dur, prev_pause, utt_start_time, mid_rep, is_rep, pmin, pmax, pmean
# 
# Used only for preprocessing:
# is_echo, is_contraction, is_oov
# 
# Used only for plotting:
# wtype, near_hes, near_part, near_rep, prange, jitter, wclass, unigram_p
# 
# Used only for modeling:
# ref, corpus, speaker, sex, ltype
# 
# Disfluency features (all binary):
# before_hes, after_hes, before_part, after_part, before_rep, after_rep, first_rep, last_rep, notlast_rep
# 
# Other binary features:
# is_open, is_disc, starts_turn
# 
# Other numeric features:
# trigram_p, unigram_p
# 
# Pronunciation features (all numeric):
# wlen_p, npron, nh, nn, fwh, fwn
# 
# Acoustic features:
# utt_rate, dur, pmin_sex, pmax_sex, pmean_sex, log_prange, imin, imax, imean, irange, log_jitter


concordance <- function(model) {
probs <- (1/(1+exp(-fitted(model))))
somers2(probs, is_err)
}

reduce <- function(v, f) {
  acc <- v[1]
  for (i in 2:length(v))
  acc <- f(acc,v[i])
  acc
}

reduce_and <- function(v) {
  reduce(v,function(x,y){x&y})
}

reduce_or <- function(v) {
  reduce(v,function(x,y){x|y})
}

#does nothing unless vector contains only 0/1 values.
binary_to_boolean <- function(v) {
  if (is.numeric(v)) {
    r<-range(v,na.rm=T)
    if ((r[1] == 0 & r[2] == 1) |
        (r[1] == 0 & r[2] == 0) |
        (r[1] == 1 & r[2] == 1)) {
      if (sum(v[v>0],na.rm=T) == sum(v[v==1],na.rm=T)) v == 1
    }
    else v
  }
  else v
}

#applies f to columns of data frame
apply_to_frame <- function(data, f, cols=c(1:ncol(data))) {
#  for (i in 1: ncol(data)) {
  for (i in cols) {
    data[i] <- f(data[[i]])
  }
  data
}

remove_feature <- function(data, feature) {
if (nrow(data) != length(feature)) {
stop("remove_feature: length of data and feature do not match!")
}
print(sprintf("Removing %1.0f items\n", sum(feature)))
data <- data[!feature,]
data <- remove_extra_levels(data)
data
}

remove_extra_levels <- function(data) {
  #remove unused levels in factors
  data$ref <- factor(data$ref)
  data$hyp <- factor(data$hyp)
  data$speaker <- factor(data$speaker)
  data$wtype <- factor(data$wtype)
data
}

#for modeling, features from acl paper only
reduce_frame_acl <- function(data) {
  attach(data)
  dd <- data.frame(index, ref_index, ref, hyp, ID, corpus, speaker, sex, starts_turn, before_hes, after_hes, before_part, after_part, before_rep, after_rep, first_rep, notlast_rep, last_rep, is_open, is_disc, wfreq, wlen_p, utt_rate, dur, pmean, log_prange, imean, irange, wfreq_2, wlen_p_2, utt_rate_2, dur_2, pmean_2, log_prange_2, imean_2, irange_2, is_err)
  detach(data)
  dd
}

#for modeling: incl. jitter, neighbor features
reduce_frame <- function(data) {
  attach(data)
  dd <- data.frame(index, ref_index, ref, hyp, ID, corpus, ltype, is_cellular, is_land, is_cordless, speaker, sex, starts_turn, before_hes, after_hes, before_part, after_part, before_rep, after_rep, first_rep, notlast_rep, last_rep, is_open, is_disc, unigram_p, trigram_p, wlen_p, npron, nh, nn, fwh, fwn, utt_rate, dur, pmean_sex, pmax_sex, pmin_sex, log_prange, imean, imax, imin, irange, log_jitter, unigram_p_2, trigram_p_2, wlen_p_2, npron_2, nh_2, fwn_2, utt_rate_2, dur_2, pmean_sex_2, pmax_sex_2, log_prange_2, imean_2, imax_2, irange_2, log_jitter_2, is_err)
  detach(data)
  dd
}

#for modeling with splines: no quads, only pmean/imean, prange/irange
reduce_frame_spline <- function(data) {
  attach(data)
  dd <- data.frame(index, ref_index, ref, hyp, ID, corpus, is_cellular, is_land, is_cordless, speaker, sex, starts_turn, before_hes, after_hes, before_part, after_part, before_rep, after_rep, notlast_rep, last_rep, is_open, is_disc, unigram_p, trigram_p, wlen_p, npron, nh, nn, fwh, fwn, utt_rate, dur, pmean_sex, log_prange, imean, irange, log_jitter, is_err)
  detach(data)
  dd
}

#for correlations: no quads, incl. all prosodic
reduce_frame_corrs <- function(data) {
  attach(data)
  dd <- data.frame(sex, starts_turn, before_hes, after_hes, before_part, after_part, before_rep, after_rep, first_rep, notlast_rep, last_rep, is_open, is_disc, unigram_p, trigram_p, wlen_p, npron, nh, nn, fwh, fwn, utt_rate, dur, pmean_sex, pmax_sex, pmin_sex, log_prange, imean, imax, imin, irange, log_jitter)
  detach(data)
  dd
}



#rescales numeric factors to have mean = 0 and a scale of 1.
rescale <- function(v) {
(v-mean(v))/(max(v)-min(v))
}

rescale_frame <- function(data) {
  for (i in 3:ncol(data)) {
    if (is.numeric(data[[i]])) {
      data[i] <- rescale(data[[i]])
    }
  }
  data
}

add_quadratic_terms <- function(data) {
data$unigram_p_2 <- data$unigram_p^2
data$trigram_p_2 <- data$trigram_p^2
data$wlen_p_2 <- data$wlen_p^2
data$utt_rate_2 <- data$utt_rate^2
data$npron_2 <- data$npron^2
data$nh_2 <- data$nh^2
data$fwn_2 <- data$fwn^2
data$dur_2 <- data$dur^2
data$pmean_sex_2 <- data$pmean_sex^2
data$pmax_sex_2 <- data$pmax_sex^2
data$imean_2 <- data$imean^2
data$imax_2 <- data$imax^2
data$log_prange_2 <- data$log_prange^2
data$irange_2 <- data$irange^2
data$log_jitter_2 <- data$log_jitter^2
data
}


#includes some non-lexical: all feats from alignment_to_features.prl
process_lexical_features <- function(data) {
  #remove failed echo-cancellation utts
  n.echo <- sum(data$is_echo)
  print(sprintf("removing %1.0f words due to echo cancellation\n", n.echo))
  data <- data[!data$is_echo,]
  #change insertion cost so that total insertion rate
  #is the same as if we counted once for each insertion
  alpha <- sum(data$eval=="i")/sum(data$wi[data$wi>0])
print(sprintf("insertion factor alpha = %1.3f\n",alpha))
  data$wi <- data$wi * alpha
  data$rer <- data$wi + data$wd + data$ws
  #now remove insertions
data <- data[data$eval != "i",]
data <- remove_extra_levels(data)
  
  #convert from 0/1 to boolean
  data <- apply_to_frame(data, binary_to_boolean)
  #fix tagging errors by hand.
  errors <- data$wclass=="error"
  data[errors & (data$ref=="whose" | data$ref=="any"),]$wclass <- "closed"
  data[errors & (data$wclass != "closed"),]$wclass <- "open"
  #remove class for hes/part.
  data[data$wclass == "hes" | data$wclass == "part",]$wclass <- NA
  # this removes extra levels with no observations:
  data$wclass <- factor(data$wclass)
  #fix unknown freqs
  data[data$trigram_p == -99,]$trigram_p <- NA
  data[data$unigram_p == -99,]$unigram_p <- NA
  #fix starts_turn
  data$starts_turn <- data$starts_turn & data$is_first

  # create new non-binary factors from binary ones
  data$near_hes <- ifelse(data$before_hes,"before", 
                          ifelse(data$after_hes,"after","none"))
  data$near_part <- ifelse(data$before_part,"before", 
                           ifelse(data$after_part,"after","none"))
  data$near_rep <- factor(rep("none",length(data$wlen_p)))
  levels(data$near_rep) <- c("before", "after", "first", "last", "middle", "none")
  data$near_rep <- ifelse(data$before_rep,"before", 
                          ifelse(data$after_rep,"after","none"))
  data$near_rep[data$is_rep] <- "middle"
  data$near_rep[data$first_rep] <- "first"
  data$near_rep[data$last_rep] <- "last"
  data$pos <- factor(rep("middle",length(data$wlen_p)))
  levels(data$pos) <- c("middle","first","last","only")
  data$pos[data$is_first] <- "first"
  data$pos[data$is_last] <- "last"
  data$pos[data$ref_length==1] <- "only"

  data$mid_rep <- data$is_rep & !data$first_rep & !data$last_rep
  data$notlast_rep <- data$is_rep & !data$last_rep

  #create new binary factors from non-binary ones
  data$is_only <- data$ref_length == 1
  data$is_open <- data$wclass == "open"
  data$is_disc <- data$wclass == "disc"
  data$is_cellular <- data$ltype == "cellular"
  data$is_cordless <- data$ltype == "cordless"
  data$is_land <- data$ltype == "land"

  #deal w/  oov words
  data$is_oov <- rep(F,length(data$wtype))
  data[is.na(data$unigram_p) & data$wtype == "full",]$is_oov <- T

data
}

#feats from running praat.
process_acoustic_features <- function(data, acoustic_data) {
  #check consistency#
  if (!reduce_and(data$ref == acoustic_data$ref))
    stop("Reference words do not match!")
# normalize pitch by sex
avgs<-tapply(acoustic_data$pmean,data$sex,mean,na.rm=T)
av_pmean <- avgs[data$sex]
  acoustic_data$pmean_sex <- acoustic_data$pmean - av_pmean
  acoustic_data$pmin_sex <- acoustic_data$pmin - av_pmean
  acoustic_data$pmax_sex <- acoustic_data$pmax - av_pmean

  acoustic_data$prange <- acoustic_data$pmax - acoustic_data$pmin
  #remove suspect intensity values
n<-nrow(acoustic_data[acoustic_data$imin<0 & !is.na(acoustic_data$imin),])
print(sprintf("removing %1.0f negative intensity values\n", n))
  acoustic_data$imin[acoustic_data$imin < 0] <- NA
  acoustic_data$irange <- acoustic_data$imax - acoustic_data$imin
  
  data <- data.frame(data,acoustic_data[,3:ncol(acoustic_data)])


  #discretize pause_dur and tally for first words only
  groups <- c(-1,0,.05,.1,.2,.3,.4,.5,1,100)
  tmp <- cut(data$pause_dur,groups, right=T)
  levels(tmp)[1] <- "0";
  levels(tmp) <- c(levels(tmp),"none")
  data$prev_pause <- tmp
  data$prev_pause[data$is_first == F] <- "none"

data$log_prange <- log(data$prange+1)
data$log_jitter <- log(data$jitter) # no zero vals.
  data
}

check_outliers <- function(data) {
op <- par(mfrow=c(5,4), mex=0.6, mai=c(.4,.4,.3,.02))
hist(data$unigram_p,30)
hist(data$trigram_p,30)
hist(data$wlen_p)
hist(data$utt_rate,30)
hist(data$dur,30)
hist(log(data$dur),30)
hist(data$pmean,30)
hist(data$pmean_sex,30)
hist(data$pmin,30)
hist(data$pmin_sex,30)
hist(data$pmax,30)
hist(data$pmax_sex,30)
hist(data$prange,30)
hist(data$log_prange,30)
hist(data$imean,30)
hist(data$imin,30)
hist(data$imax,30)
hist(data$irange,30)
hist(data$jitter,30)
hist(data$log_jitter,30)
par(op)
}

remove_outliers <- function(data) {
l<-length(data[data$utt_rate > 27, 1])
data<-data[data$utt_rate <= 27,]
l<- l+length(data[data$dur > 1.25, 1])
data<-(data[data$dur <= 1.25,])
l<- l+length(data[data$log_jitter < -7, 1])
data<-(data[data$log_jitter >= -7,])
print(sprintf("removed %1.0f outliers\n", l))
data
}

full_words_only <- function(data) {
print("Removing OOV")
#remove oov words
tmp <- remove_feature(data, data$is_oov)
#remove all but full words
full <- remove_feature(tmp, tmp$wtype != "full")
full
}

prosodic_data_only <- function(data) {
#mportant to remove NA values before rescaling or calling discretize
prosodic <- remove_feature(data, 
	is.na(data$irange) | is.na(data$prange) |
	is.na(data$log_jitter))
prosodic
}

remove_contractions <- function(data) {
print("Removing contractions")
data <- remove_feature(data, data$is_contraction)
data
}

acoustic_data<-read.table("features/acoustic.dat",header = TRUE);

#full sri set
sri_data<-read.table("features/sri_lex.dat",header = TRUE);
sri_data <- process_lexical_features(sri_data)
sri_data <- process_acoustic_features(sri_data,acoustic_data)
sri_full <- full_words_only(sri_data)
sri_full_prosodic <- prosodic_data_only(sri_full)
sri_prosodic <- remove_contractions(sri_full_prosodic)
sri_rescaled <- reduce_frame(add_quadratic_terms(rescale_frame(remove_outliers(sri_prosodic))))

#full cambridge set
cam_data<-read.table("features/cambridge_lex.dat",header = TRUE);
cam_data <- process_lexical_features(cam_data)
cam_data <- process_acoustic_features(cam_data,acoustic_data)
cam_full <- full_words_only(cam_data)
cam_full_prosodic <- prosodic_data_only(cam_full)
cam_prosodic <- remove_contractions(cam_full_prosodic)
cam_rescaled <- reduce_frame(add_quadratic_terms(rescale_frame(remove_outliers(cam_prosodic))))


#check_outliers(sri_full)
